#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools_scm

extensions = [
    'sphinx.ext.autodoc',
]

# General information about the project.
project = 'jaraco.mongodb'
copyright = '2015 Jason R. Coombs'

import os
root = os.path.relpath(os.path.join(os.path.dirname(__file__), '..'))

# The short X.Y version.
version = setuptools_scm.get_version(root=root)
# The full version, including alpha/beta/rc tags.
release = version

master_doc = 'index'



import datetime
import dateutil

import py.test

py.test.importorskip("cherrypy")

from jaraco.mongodb import sessions


@py.test.fixture(scope='function')
def database(request, mongodb_instance):
    """
    Return a MongoDB database suitable for testing auth. Remove the
    collection between every test.
    """
    database = mongodb_instance.get_connection().sessions_test
    request.addfinalizer(database.sessions.remove)
    return database

class TestSessions(object):
    def test_time_conversion(self):
        local_time = datetime.datetime.now().replace(
            microsecond=0)
        local_time = sessions.Session._make_aware(local_time)
        utc_aware = datetime.datetime.utcnow().replace(
            tzinfo=dateutil.tz.tzutc(),
            microsecond=0)
        assert local_time == utc_aware

    def test_time_conversion2(self):
        local_time = datetime.datetime.now().replace(microsecond=0)
        round_local = sessions.Session._make_local(sessions.Session._make_utc(
            local_time))
        assert round_local == local_time
        assert round_local.tzinfo is None

    def test_session_persists(self, database):
        session = sessions.Session(None, database = database)
        session['x'] = 3
        session['y'] = "foo"
        session.save()
        session_id = session.id
        del session
        session = sessions.Session(session_id, database=database)
        assert session['x'] == 3
        assert session['y'] == 'foo'

    def test_locked_session(self, database):
        session = sessions.Session(None, database=database)
        session.acquire_lock()
        session['x'] = 3
        session['y'] = "foo"
        session.save()
        session_id = session.id
        del session
        session = sessions.Session(session_id, database=database)
        assert session['x'] == 3

    @py.test.mark.xfail
    def test_numeric_keys(self, database):
        session = sessions.Session(None, database=database, use_modb=True)
        session.acquire_lock()
        session[3] = 9
        session.save()
        session_id = session.id
        del session
        session = sessions.Session(session_id, database=database, use_modb=True)
        assert 3 in session
        assert session[3] == 9



import os
import sys
import tempfile
import subprocess
import glob
import collections
import importlib
import shutil
import functools
import logging

import portend
from jaraco.services import paths
from jaraco import services


log = logging.getLogger(__name__)


class MongoDBFinder(paths.PathFinder):
    windows_installed = glob.glob('/Program Files/MongoDB/Server/???/bin')
    windows_paths = [
        # symlink Server/current to Server/X.X
        '/Program Files/MongoDB/Server/current/bin',
        # symlink MongoDB to mongodb-win32-x86_64-2008plus-X.X.X-rcX
        '/Program Files/MongoDB/bin',
    ] + list(reversed(windows_installed))
    heuristic_paths = [
        # on the path
        '',
        # 10gen Debian package
        '/usr/bin',
        # custom install in /opt
        '/opt/mongodb/bin',
    ] + windows_paths

    # allow the environment to stipulate where mongodb must
    #  be found.
    env_paths = [
        os.path.join(os.environ[key], 'bin')
        for key in ['MONGODB_HOME']
        if key in os.environ
    ]
    candidate_paths = env_paths or heuristic_paths
    exe = 'mongod'
    args = ['--version']

    @classmethod
    def find_binary(cls):
        return os.path.join(cls.find_root(), cls.exe)


class MongoDBService(MongoDBFinder, services.Subprocess, services.Service):
    port = 27017

    @services.Subprocess.PortFree()
    def start(self):
        super(MongoDBService, self).start()
        # start the daemon
        mongodb_data = os.path.join(sys.prefix, 'var', 'lib', 'mongodb')
        cmd = [
            self.find_binary(),
            '--dbpath=' + mongodb_data,
        ]
        self.process = subprocess.Popen(cmd, stdout=self.get_log())
        self.wait_for_pattern('waiting for connections on port (?P<port>\d+)')
        log.info('%s listening on %s', self, self.port)

is_virtualenv = lambda: hasattr(sys, 'real_prefix')

class MongoDBInstance(MongoDBFinder, services.Subprocess, services.Service):
    @staticmethod
    def get_data_dir():
        data_dir = None
        if is_virtualenv():
            # use the virtualenv as a base to store the data
            data_dir = os.path.join(sys.prefix, 'var', 'data')
            if not os.path.isdir(data_dir):
                os.makedirs(data_dir)
        return tempfile.mkdtemp(dir=data_dir)

    def start(self):
        super(MongoDBInstance, self).start()
        self.data_dir = self.get_data_dir()
        if not hasattr(self, 'port') or not self.port:
            self.port = self.find_free_port()
        cmd = [
            self.find_binary(),
            '--dbpath', self.data_dir,
            '--port', str(self.port),
            '--noprealloc',
            '--nojournal',
            '--nohttpinterface',
            '--syncdelay', '0',
            '--ipv6',
            '--noauth',
            '--setParameter', 'textSearchEnabled=true',
        ]
        if hasattr(self, 'bind_ip'):
            cmd.extend(['--bind_ip', self.bind_ip])
        self.process = subprocess.Popen(cmd, stdout=self.get_log())
        portend.occupied('localhost', self.port, timeout=3)
        log.info('{self} listening on {self.port}'.format(**locals()))

    def get_connection(self):
        pymongo = importlib.import_module('pymongo')
        return pymongo.MongoClient('localhost', self.port)

    def get_connect_hosts(self):
        return ['localhost:{self.port}'.format(**locals())]

    def get_uri(self):
        return 'mongodb://' + ','.join(self.get_connect_hosts())

    def stop(self):
        super(MongoDBInstance, self).stop()
        shutil.rmtree(self.data_dir)


class MongoDBReplicaSet(MongoDBFinder, services.Service):
    replica_set_name = 'test'

    def start(self):
        super(MongoDBReplicaSet, self).start()
        self.data_root = tempfile.mkdtemp()
        self.instances = list(map(self.start_instance, range(3)))
        # initialize the replica set
        self.instances[0].connect().admin.command(
            'replSetInitiate', self.build_config())
        # wait until the replica set is initialized
        get_repl_set_status = functools.partial(
            self.instances[0].connect().admin.command, 'replSetGetStatus', 1
        )
        errors = importlib.import_module('pymongo.errors')
        log.info('Waiting for replica set to initialize')
        while True:
            try:
                res = get_repl_set_status()
                if res.get('myState') != 1: continue
            except errors.OperationFailure:
                continue
            break

    def start_instance(self, number):
        port = self.find_free_port()
        data_dir = os.path.join(self.data_root, 'r{number}'.format(**locals()))
        os.mkdir(data_dir)
        cmd = [
            self.find_binary(),
            '--replSet', self.replica_set_name,
            '--noprealloc',
            '--smallfiles',
            '--oplogSize', '10',
            '--dbpath', data_dir,
            '--port', str(port),
        ]
        log_file = self.get_log(number)
        process = subprocess.Popen(cmd, stdout=log_file)
        portend.occupied('localhost', port, timeout=50)
        log.info('{self}:{number} listening on {port}'.format(**locals()))
        return InstanceInfo(data_dir, port, process, log_file)

    def get_log(self, number):
        log_name = 'r{number}.log'.format(**locals())
        log_filename = os.path.join(self.data_root, log_name)
        log_file = open(log_filename, 'a')
        return log_file

    def is_running(self):
        return hasattr(self, 'instances') and all(
            instance.process.returncode is None for instance in self.instances)

    def stop(self):
        super(MongoDBReplicaSet, self).stop()
        for instance in self.instances:
            if instance.process.returncode is None:
                instance.process.terminate()
                instance.process.wait()
            instance.log_file.close()
        del self.instances
        shutil.rmtree(self.data_root)

    def build_config(self):
        return dict(
            _id = self.replica_set_name,
            members = [
                dict(
                    _id=number,
                    host='localhost:{instance.port}'.format(**locals()),
                ) for number, instance in enumerate(self.instances)
            ]
        )

    def get_connect_hosts(self):
        return ['localhost:{instance.port}'.format(**locals())
            for instance in self.instances]

InstanceInfoBase = collections.namedtuple('InstanceInfoBase',
    'path port process log_file')
class InstanceInfo(InstanceInfoBase):
    def connect(self):
        hp = 'localhost:{self.port}'.format(**locals())
        pymongo = __import__('pymongo')
        rp = pymongo.ReadPreference.PRIMARY_PREFERRED
        return pymongo.MongoClient(hp, read_preference=rp)



import six


def project(*args, **kwargs):
    """
    Build a projection for MongoDB.

    Due to https://jira.mongodb.org/browse/SERVER-3156, until MongoDB 2.6,
    the values must be integers and not boolean.

    >>> project(a=True) == {'a': 1}
    True

    Once MongoDB 2.6 is released, replace use of this function with a simple
    dict.
    """
    projection = dict(*args, **kwargs)
    return {key: int(value) for key, value in six.iteritems(projection)}


def compat_explain(cur):
    """
    Simulate MongoDB 3.0 explain result on prior versions.
    http://docs.mongodb.org/v3.0/reference/explain-results/
    """
    res = cur.explain()
    if 'nscannedObjects' in res:
        res['executionStats'] = dict(
            nReturned=res.pop('n'),
            totalKeysExamined=res.pop('nscanned'),
            totalDocsExamined=res.pop('nscannedObjects'),
            executionTimeMillis=res.pop('millis'),
        )
    return res



"""
A MongoDB-backed CherryPy session store.

Although this module requires CherryPy, it does not impose the requirement
on the package, as any user of this module will already require CherryPy.

To enable these sessions, your code must call :meth:`Session.install()` and
then configure the CherryPy endpoint to use MongoDB sessions. For example::

	jaraco.mongodb.sessions.Session.install()

	session_config = {
		'sessions.on': True,
		'sessions.storage_type': 'MongoDB',
		'sessions.database': pymongo.MongoClient().database,
	}
	config = {
		'/': session_config,
	}

	cherrypy.quickstart(..., config=config)

The ``jaraco.modb`` module implements the codec interface, so may be used
to encode more complex objects in the session::

	session_config.update({
		'sessions.codec': jaraco.modb,
	})

"""

import datetime
import time
import logging
import pprint

import pymongo.errors
import cherrypy
import dateutil.tz

from . import timers

log = logging.getLogger(__name__)

class LockTimeout(RuntimeError): pass

class NullCodec(object):
	def decode(self, data):
		return data

	def encode(self, data):
		return data

class Session(cherrypy.lib.sessions.Session):
	"""
	A MongoDB-backed CherryPy session store. Takes the following params:

		database: the pymongo Database object.
		collection_name: The name of the collection to use in the db.
		codec: An object with 'encode' and 'decode' methods, used to encode
			objects before saving them to MongoDB and decode them when loaded
			from MongoDB.
		lock_timeout: A timedelta or numeric seconds indicating how long
			to block acquiring a lock. If None (default), acquiring a lock
			will block indefinitely.
	"""

	codec = NullCodec()
	"by default, objects are passed directly to MongoDB"

	def __init__(self, id, **kwargs):
		kwargs.setdefault('collection_name', 'sessions')
		kwargs.setdefault('lock_timeout', None)
		super(Session, self).__init__(id, **kwargs)
		self.setup_expiration()
		if isinstance(self.lock_timeout, (int, float)):
			self.lock_timeout = datetime.timedelta(seconds=self.lock_timeout)
		if not isinstance(self.lock_timeout, (datetime.timedelta, type(None))):
			raise ValueError("Lock timeout must be numeric seconds or "
				"a timedelta instance.")

	@classmethod
	def install(cls):
		"""
		Add this session to the cherrypy session handlers. CherryPy looks
		for session classes in vars(cherrypy.lib.sessions) with a name
		in title-case followed by "Session".
		"""
		cherrypy.lib.sessions.MongodbSession = cls

	@property
	def collection(self):
		return self.database[self.collection_name]

	def setup_expiration(self):
		"""
		Use pymongo TTL index to automatically expire sessions.
		"""
		self.collection.ensure_index('_expiration_datetime',
			expireAfterSeconds=0)

	def _exists(self):
		return bool(self.collection.find_one(self.id))

	def _load(self):
		filter = dict(
			_id=self.id,
			_expiration_datetime={'$exists': True},
		)
		projection=dict(_id=False)
		doc = self.collection.find_one(filter, projection)
		if not doc: return
		expiration_time = doc.pop('_expiration_datetime')
		doc = self.codec.decode(doc)
		return (doc, self._make_local(expiration_time))

	@staticmethod
	def _make_aware(local_datetime):
		return local_datetime.replace(tzinfo=dateutil.tz.tzlocal())

	@staticmethod
	def _make_utc(local_datetime):
		return Session._make_aware(local_datetime).astimezone(
			dateutil.tz.tzutc()
		)

	@staticmethod
	def _make_local(utc_datetime):
		"""
		For a naive utc_datetime, return the same time in the local timezone
		(also naive).
		"""
		return utc_datetime.replace(
			tzinfo = dateutil.tz.tzutc()
		).astimezone(dateutil.tz.tzlocal()).replace(
			tzinfo = None
		)

	def _save(self, expiration_datetime):
		data = dict(self._data)
		data = self.codec.encode(data)
		# CherryPy defines the expiration in local time, which may be
		#  different for some hosts. Convert it to UTC before sticking
		#  it in the database.
		expiration_datetime = self._make_utc(expiration_datetime)
		data.update(
			_expiration_datetime = expiration_datetime,
			_id = self.id,
		)
		try:
			self.collection.save(data)
		except pymongo.errors.InvalidDocument:
			log.warning("Unable to save session:\n%s",
				pprint.pformat(data))
			raise

	def _delete(self):
		self.collection.remove(self.id)

	def acquire_lock(self):
		"""
		Acquire the lock. Blocks indefinitely until lock is available
		unless `lock_timeout` was supplied. If the lock_timeout elapses,
		raises LockTimeout.
		"""
		# first ensure that a record exists for this session id
		try:
			self.collection.insert(dict(_id=self.id))
		except pymongo.errors.DuplicateKeyError:
			pass
		unlocked_spec = dict(_id=self.id, locked=None)
		lock_timer = (
			timers.Timer.after(self.lock_timeout)
			if self.lock_timeout
			else timers.NeverExpires()
		)
		while not lock_timer.expired():
			locked_spec = {'$set': dict(locked=datetime.datetime.utcnow())}
			res = self.collection.update(unlocked_spec, locked_spec)
			if res['updatedExisting']:
				# we have the lock
				break
			time.sleep(0.1)
		else:
			raise LockTimeout("Timeout acquiring lock for {self.id}"
				.format(**vars()))
		self.locked = True

	def release_lock(self):
		record_spec = dict(_id=self.id)
		self.collection.update(record_spec, {'$unset': {'locked': 1}})
		# if no data was saved (no expiry), remove the record
		record_spec.update(_expiration_datetime={'$exists': False})
		self.collection.remove(record_spec)
		self.locked = False

	def __len__(self):
		return self.collection.count()
